{-# LANGUAGE PatternGuards, TypeOperators #-}
module Opt where
import qualified Data.Set as S

import AST
import Label
import Type

optimise :: TypedTerm ::: Type -> LabelGen (TypedTerm ::: Type)
optimise (TypedTerm term ::: typ) = case term of
  Var v      -> return $ TypedTerm (Var v) ::: typ
  Abs v mt t -> do
    t' <- optimise t
    case t' of
      (TypedTerm (t_inner :$ (TypedTerm (Var v') ::: _)) ::: _)
        | v == v', v `S.notMember` fv t_inner -> return t_inner
      _ -> return $ TypedTerm (Abs v mt t') ::: typ
  t1 :$ t2 -> do
    t1' <- optimise t1
    t2' <- optimise t2
    case (t1', t2') of
      -- Desugar if
      ( TypedTerm ((TypedTerm ((TypedTerm (Const c) ::: _) :$ b) ::: _) :$ (TypedTerm t ::: _)) ::: _
       , TypedTerm f ::: typ1 :=>: typ2)
        | cName c == "if then else" -> do
        l <- newLabel "$x"
        let x = show l 
            ite = Const c .:: (tint :=>: typ2 :=>: typ2 :=>: typ2)
            iteb = (ite :$ b) .:: (typ2 :=>: typ2 :=>: typ2)
            t' = ((t .:: (typ1 :=>: typ2)) :$ (Var x .:: typ1)) .:: typ2
            f' = ((f .:: (typ1 :=>: typ2)) :$ (Var x .:: typ1)) .:: typ2
            itebt' = (iteb :$ t') .:: (typ2 :=>: typ2)
            itebt'f' = (itebt' :$ f') .:: typ2
        optimise $ Abs x (Just typ1) itebt'f' .:: typ
      ( TypedTerm ((TypedTerm ((TypedTerm (Const c) ::: _) :$ b) ::: _) :$ (TypedTerm t ::: _)) ::: _
       , TypedTerm f ::: typ1 :*: typ2)
        | cName c == "if then else" -> do
        let ite1 = Const c .:: tint :=>: typ1 :=>: typ1 :=>: typ1
            ite2 = Const c .:: tint :=>: typ2 :=>: typ2 :=>: typ2
            iteb1 = (ite1 :$ b) .:: (typ1 :=>: typ1 :=>: typ1)
            iteb2 = (ite2 :$ b) .:: (typ2 :=>: typ2 :=>: typ2)
            p1 = Proj1 .:: ((typ1 :*: typ2) :=>: typ1)
            p2 = Proj2 .:: ((typ1 :*: typ2) :=>: typ2)
            ti1 = (p1 :$ (t .:: (typ1 :*: typ2))) .:: typ1
            ti2 = (p2 :$ (t .:: (typ1 :*: typ2))) .:: typ2
            f1 = (p1 :$ (f .:: (typ1 :*: typ2))) .:: typ1
            f2 = (p2 :$ (f .:: (typ1 :*: typ2))) .:: typ2
            itebtf1 = (((iteb1 :$ ti1) .:: (typ1 :=>: typ1)) :$ f1) .:: typ1
            itebtf2 = (((iteb2 :$ ti2) .:: (typ2 :=>: typ2)) :$ f2) .:: typ2
        optimise (((((Pair .:: (typ1 :=>: typ2 :=>: typ1 :*: typ2)) :$ itebtf1) .:: (typ2 :=>: typ1 :*: typ2)) :$ itebtf2) .:: (typ1 :*: typ2))
      -- (\x. t) t', x unused --> t
      (TypedTerm (Abs v (Just _) t) ::: _, _) | v `S.notMember` fv t -> return t
      ( TypedTerm (Abs v (Just vtyp) (TypedTerm ((TypedTerm ta1 ::: typ1) :$ (TypedTerm ta2 ::: typ2)) ::: _)) ::: _
       , TypedTerm t ::: ttyp)
        -- (\x. t1 t2) t, x unused in t1 --> t1 ((\x. t2) t)
        | v `S.notMember` fv ta1 ->
          optimise $ ((ta1 .:: typ1) :$ ((Abs v (Just vtyp) (ta2 .:: typ2) .:: vtyp :=>: typ2) :$ (t .:: ttyp) .:: typ2)) .:: typ
        {- this breaks if then else
        -- (\x. t1 t2) t, x unused in t2 --> ((\x. t1) t) t2
        | v `S.notMember` fv t2 ->
          optimise $ (((Abs v (Just vtyp) (t1 .:: typ1) .:: vtyp :=>: typ1) :$ (t .:: ttyp) .:: typ1) :$ (t2 .:: typ2)) .:: typ
        -}
      ( TypedTerm Proj1 ::: _
       ,TypedTerm ((TypedTerm ((TypedTerm Pair ::: _) :$ tp1) ::: _) :$ _tp2) ::: _)
        -> return tp1
      ( TypedTerm Proj2 ::: _
       ,TypedTerm ((TypedTerm ((TypedTerm Pair ::: _) :$ _tp1) ::: _) :$ tp2) ::: _)
        -> return tp2
      -- Desugar seq
      ( TypedTerm ((TypedTerm (Const c) ::: (_ :=>: _ :=>: (a :*: b))) :$ ts1) ::: _
       ,ts2)
        | cName c == ";" -> do
        let seq1 = Const c .:: tcom :=>: a :=>: a
            seq2 = Const c .:: tcom :=>: b :=>: b
            seq1ts1 = (seq1 :$ ts1) .:: (a :=>: a)
            seq2ts1 = (seq2 :$ ts1) .:: (b :=>: b)
            fstts2 = ((Proj1 .:: (a :*: b :=>: a)) :$ ts2) .:: a
            sndts2 = ((Proj2 .:: (a :*: b :=>: b)) :$ ts2) .:: b
            seq1ts1ts2 = (seq1ts1 :$ fstts2) .:: a
            seq2ts1ts2 = (seq2ts1 :$ sndts2) .:: b
        optimise (((((Pair .:: a :=>: b :=>: a :*: b) :$ seq1ts1ts2) .:: b :=>: a :*: b) :$ seq2ts1ts2) .:: a :*: b)
      ( TypedTerm ((TypedTerm (Const c) ::: (_ :=>: _ :=>: (a :=>: b))) :$ ts1) ::: _
       ,ts2)
        | cName c == ";" -> do
        l <- newLabel "$x"
        let x = show l
            seq1 = Const c .:: (tcom :=>: b :=>: b)
            seq1ts1 = (seq1 :$ ts1) .:: (b :=>: b)
            ts2' = (ts2 :$ (Var x .:: a)) .:: b

            seq1ts1ts2' = (seq1ts1 :$ ts2') .:: b
        optimise $ Abs x (Just a) seq1ts1ts2' .:: (a :=>: b)
      _ -> return $ (t1' :$ t2') .:: typ

  t :@ n -> do
    t' <- optimise t
    return $ t' :@ n .:: typ
  _ -> return $ term .:: typ
  where
    infix 2 .::
    tt .:: typp = TypedTerm tt ::: typp
