module Main where

import Control.Arrow(first)
import Control.Monad.Trans
import Data.List
import Data.Monoid
import Language.C.Generate
import System.FilePath

import Component
import Compiler
import ComponentCompiler
import ParseToAST
import TypeChecker


prog1 = "5"
prog2 = "5 + 6"
prog3 = "(\\x. x) 5"
prog4 = "(\\x. x + 4 + x + x) 7"
prog5 = "(\\x y. x + y) 2 3"
prog6 = "((\\f. f 3 + f 4 + f 5)@a (\\x. x + 1))@b"
prog7 = "(fix (\\x. 5))@a"
prog8 = "(fix (\\(f : Int -> Int) (x : Int). if x then (1 + (f (x - 1))) else 2) 1)@a"
prog9 = "(fix (\\f x. if x then x + (f (x - 1)) else 0) 10)@a"


compileTerm s filename =
  let pccs = do
        (Port MsgInt p, _) <- compile $ typeCheckOuter $ parseTerm s
        emit =<< lift (Component.print p)
      (ccns, ps)    = snd $ evalCompiler pccs
      (ccs, ns)     = unzip ccns
      connectedccs = zip (connectPorts ps ccs) ns
      combinedccs  = combineComponents connectedccs
  in do
    let outdir = "out"
        bindir = "bin"
        cc     = "mpicc"
        run    = "mpirun"
    Prelude.print ps
    sequence_ [putStrLn $ "#Node " ++ show n ++ ": " ++ prettyComponentCode cc | (cc, n) <- connectedccs]
    putStrLn "################"
    sequence_ [putStrLn $ "#Node " ++ n ++ ": " ++ prettyComponentCode cc | (cc, n) <- combinedccs]
    files <- sequence
      [ do let outfile = filename ++ nodeid
               cfile   = outfile <.> "c"
               runfile = bindir </> outfile <.> "out"
           writeFile (outdir </> cfile) (generate code)
           return (cfile, runfile)
      | (nodeid, code) <- compileComponents combinedccs]
    let compilelines = [unwords [cc, cfile, {-"-O2",-} "-g", "-o", runfile]  | (cfile, runfile) <- files]
        runlines     = [unwords ["-np", "1", runfile] | (_, runfile) <- files]
        makefile = unlines $ concat $
          [ ["all:"]
          , indented [ unwords ["mkdir", "-p", bindir]
                     , unwords ["cp", "../prelude.h", "."]
                     ]
          , indented compilelines
          , ["run:"]
          , indented [unwords [run, intercalate " : " runlines]]
          , ["clean:"]
          , indented [unwords ["$(RM)", bindir </> "*"]]
          ]
    writeFile (outdir </> "Makefile") makefile

  where indented = map ("\t" ++)
