{-# LANGUAGE DeriveDataTypeable, DoRec #-}
module Parser where

import Control.Applicative
import Control.Arrow
import Data.Data

import Data.Parser.Grempa.Grammar

import Lexer

data PType
  = PTInt
  | PTFun PType PType
  deriving (Eq, Show, Typeable)

data Expr
  = EApp Expr Expr
  | EOp  String
  | EVar String
  | ENum Integer
  | ELam [(String, Maybe PType)] Expr
  | EAt Expr String
  | EFix
  -- | ECase Expr [(Integer, Expr)] Expr
  | EIf
  deriving (Eq, Show, Typeable)

fun :: Grammar Tok Expr
fun = do
  rec
    v <- rule [fromTok <@> var]
    vars <- several v
    typ <- levels $ do
      rec
        typ  <- lrule [ PTFun <@> typ1 <# RightArrow <#> typ ]
        typ1 <- lrule [ PTInt <@ con
                      , paren typ ]
      return typ
    vtyped   <- rule [typed <@> v <# Colon <#> typ]
    vuntyped <- rule [untyped <@> v]
    vptyped  <- rule [paren vtyped]
    vmtype   <- rule [ id <@> vptyped
                     , id <@> vuntyped
                     ]
    vmtypes   <- several vmtype
    vtypes  <- rule [ id <@> vmtypes
                    , (:[]) <@> vtyped
                    ]
    branch <- rule [mkbranch <@> num <# RightArrow <#> expr <# SemiColon
                   ]
    branches <- several branch

    expr <- levels $ do
        rec
          expr  <- lrule
              [ ELam <@ Lambda <#> vtypes <# Dot <#> expr
              ]
          expr1 <- lrule
              [ eif <@ If <#> expr2 <# Then <#> expr <# Else <#> expr]
              --[ ECase <@> expr <# Of <# LCurl <#> branches
                -- <# Underscore <# RightArrow <#> expr <# RCurl ]
          expr2 <- lrule
              [ eop <@> expr2 <#> op <#> expr3
              ]
          expr3 <- lrule
              [ EApp  <@> expr3 <#> expr4
              ]
          expr4 <- lrule
              [ EAt <@> expr5 <# At <#> v ]
          expr5 <- lrule
              [ EVar <@> v
              , EFix <@ Fix
              , ENum . fromNum <@> num
              , paren expr
              --EVar . fromTok <@ LParen <#> op <# RParen
              ]
        return expr

  return expr
  where
    mkbranch n e = (fromNum n, e)
    typed var typ = (var, Just typ)
    untyped var   = (var, Nothing)
    paren x = id <@ LParen <#> x <# RParen
    eop a o b = (EOp (fromTok o) `EApp` a) `EApp` b
    eif a b c = ((EIf `EApp` a) `EApp` b) `EApp` c
