{-# LANGUAGE TypeOperators #-}
module TypeChecker where

import Control.Applicative
import Control.Arrow(first, second)
import Control.Monad.State
import Data.Map(Map)
import qualified Data.Map as M
import Data.Maybe
import Data.Monoid

import AST

type TypeChecker a = State [Var] a
evalTypeChecker :: TypeChecker a -> a
evalTypeChecker = flip evalState $ map (("t" ++) . show) [0..]

newTVar :: TypeChecker Type
newTVar = do
  v:vs <- get
  put vs
  return (TVar v)

type Cxt = Map Var Type

typeCheckOuter :: Term -> TypedTerm ::: Type
typeCheckOuter t = case typ' of
    TInt -> tt' ::: TInt -- error $ show (typ, tt', subst)
    _    -> error $ "Term doesn't have type Int, it has type: " ++ show typ
  where
    (ttt, subst) = evalTypeChecker $ typeCheck t mempty
    tt  ::: typ  = typeSubst subst ttt
    tt' ::: typ' = typeSubst (unify typ TInt) (tt ::: typ)

typeCheck :: Term -> Cxt -> TypeChecker (TypedTerm ::: Type, TSubst)
typeCheck (Term term) cxt = case term of
  Var v            -> do
    let typ = fromMaybe (error $ "Not in scope: " ++ v) $ M.lookup v cxt
    return (Var v .:: typ, mempty)
  Abs v mtype1 t -> do
    type1 <- maybe newTVar return mtype1
    let cxt' = M.insert v type1 cxt
    (tt ::: type2, subst) <- typeCheck t cxt'
    return (Abs v (Just type1) (tt ::: type2) .:: type1 :->: type2, subst)
  t1 :$ t2         -> do
    restype <- newTVar
    (tt1@(_ ::: funtype), subst1) <- typeCheck t1 cxt
    (tt2@(_ ::: argtype), subst2) <- typeCheck t2 (typeSubst subst1 cxt)
    let subst3  = uncurry unify $ typeSubst (mconcat [subst2, subst1])
                                    (funtype, (argtype :->: restype))
    let subst = mconcat [subst3, subst2, subst1]
    return (tt1 :$ tt2 .:: restype, subst)
  Lit n            -> return (Lit n .:: TInt, mempty)
  Plus             -> return (Plus .:: TInt :->: TInt :->: TInt, mempty)
  Times            -> return (Times .:: TInt :->: TInt :->: TInt, mempty)
  Minus            -> return (Minus .:: TInt :->: TInt :->: TInt, mempty)
  Fix              -> do
    typ <- newTVar
    return (Fix .:: (typ :->: typ) :->: typ, mempty)
  If              -> do
    return (If .:: TInt :->: TInt :->: TInt :->: TInt, mempty)
  t :@ n           -> do
    (tt@(_ ::: typ), subst) <- typeCheck t cxt
    return (tt :@ n .:: typ, subst)
  where
    infix 2 .::
    tt .:: typ = TypedTerm tt ::: typ
